@@ -4,17 +4,23 @@
 from __future__ import annotations
 
 import struct
-from typing import Any, Callable, NamedTuple, Union
+from collections.abc import Callable
+from typing import Any, NamedTuple
 
 from PIL import Image
 from PIL.TiffTags import TAGS_V2 as TIFF_TAGS
 
-BytesLike = Union[bytes, memoryview]
-MutableBytesLike = Union[bytearray, memoryview]
+
+class ImageDecompressionError(Exception):
+    """Image decompression error."""
+
+
+BytesLike = bytes | memoryview
+MutableBytesLike = bytearray | memoryview
 
 
 def _next_multiple(n: int, k: int) -> int:
-    """Returns the multiple of k that is greater than or equal n.
+    """Return the multiple of k that is greater than or equal n.
 
     >>> _next_multiple(101, 4)
     104
@@ -79,6 +85,9 @@ def _2bit_inner_loop(in_: BytesLike, out: MutableBytesLike, scale: int) -> None:
     """Unpack 2-bit values to their 8-bit equivalents.
 
     Thus *out* must be 4x at long as *in*.
+
+    Images of this type are quite rare in practice, so we don't
+    optimize this loop.
     """
     for n, val in enumerate(in_):
         out[4 * n] = int((val >> 6) * scale)
@@ -91,6 +100,9 @@ def _4bit_inner_loop(in_: BytesLike, out: MutableBytesLike, scale: int) -> None:
     """Unpack 4-bit values to their 8-bit equivalents.
 
     Thus *out* must be 2x at long as *in*.
+
+    Images of this type are quite rare in practice, so we don't
+    optimize this loop.
     """
     for n, val in enumerate(in_):
         out[2 * n] = int((val >> 4) * scale)
@@ -104,7 +116,23 @@ def image_from_byte_buffer(buffer: BytesLike, size: tuple[int, int], stride: int
     with odd image widths.
     """
     ystep = 1  # image is top to bottom in memory
-    return Image.frombuffer('L', size, buffer, "raw", 'L', stride, ystep)
+    # Even if the image is type 'P' (palette), we create it as a 'L' grayscale
+    # at this step. The palette is attached later.
+    try:
+        return Image.frombuffer('L', size, buffer, "raw", 'L', stride, ystep)
+    except ValueError as e:
+        if 'buffer is not large enough' in str(e):
+            # If Pillow says the buffer is not large enough, then we're going
+            # to guess that it's padded to a multiple of 4 bytes. In practice
+            # the image may just be corrupted.
+            try:
+                return Image.frombuffer(
+                    'L', size, buffer, "raw", 'L', (size[0] + 3) // 4, ystep
+                )
+            except ValueError as e2:
+                raise ImageDecompressionError(str(e2)) from e2
+        else:
+            raise ImageDecompressionError() from e
 
 
 def _make_rgb_palette(gray_palette: bytes) -> bytes:
@@ -134,12 +162,6 @@ def image_from_buffer_and_palette(
     1/2/4-bit images must be unpacked (no scaling!) to byte buffers first, such
     that every 8-bit integer is an index into the palette.
     """
-
-    # Reminder Pillow palette byte order unintentionally changed in 8.3.0
-    # https://github.com/python-pillow/Pillow/issues/5595
-    # 8.2.0: all aligned by channel (very nonstandard)
-    # 8.3.0: all channels for one color followed by the next color (e.g. RGBRGBRGB)
-
     if base_mode == 'RGB':
         im = image_from_byte_buffer(buffer, size, stride)
         im.putpalette(palette, rawmode=base_mode)
@@ -165,8 +187,10 @@ def fix_1bit_palette_image(
     im = im.convert('P')
     if base_mode == 'RGB' and len(palette) == 6:
         # rgbrgb -> rgb000000...rgb
-        palette = palette[0:3] + (b'\x00\x00\x00' * (256 - 2)) + palette[3:6]
-        im.putpalette(palette, rawmode='RGB')
+        expanded_palette = b''.join(
+            [palette[0:3], (b'\x00\x00\x00' * (256 - 2)), palette[3:6]]
+        )
+        im.putpalette(expanded_palette, rawmode='RGB')
     elif base_mode == 'L':
         try:
             im.putpalette(palette, rawmode='L')
@@ -179,11 +203,14 @@ def fix_1bit_palette_image(
 
 def generate_ccitt_header(
     size: tuple[int, int],
+    *,
     data_length: int,
     ccitt_group: int,
+    t4_options: int | None,
     photometry: int,
     icc: bytes,
 ) -> bytes:
+    """Generate binary CCITT header for image with given parameters."""
     tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H'
 
     tag_keys = {tag.name: key for key, tag in TIFF_TAGS.items()}  # type: ignore
@@ -215,7 +242,10 @@ def generate_ccitt_header(
     add_ifd('ImageLength', height)
     add_ifd('BitsPerSample', 1)
     add_ifd('Compression', ccitt_group)
-    add_ifd('PhotometricInterpretation', int(photometry))
+    add_ifd('FillOrder', 1)
+    if t4_options is not None:
+        add_ifd('T4Options', t4_options)
+    add_ifd('PhotometricInterpretation', photometry)
     add_ifd('StripOffsets', lambda: image_offset)
     add_ifd('RowsPerStrip', height)
     add_ifd('StripByteCounts', data_length)
